# -*- coding: utf-8 -*-

"""
.. module:: dict.decoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Decodes a term from a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import arrow

from pyessv._model import Term
from pyessv._model import Authority
from pyessv._model import Scope
from pyessv._model import Collection
from pyessv._utils.compat import str



def decode(obj):
    """Decodes a term from a dictionary.

    :param dict obj: Dictionary to be decoded.

    :returns: Decoded term.
    :rtype: pyessv.Term

    """
    typeof = _get_type(obj)
    instance = typeof()
    _decode_node(obj, instance)
    _DECODERS[typeof](obj, instance)

    return instance


def _get_type(obj):
    """Returns type to be decoded.

    """
    try:
        typekey = obj['_type']
    except KeyError:
        raise TypeError('Decoding type key not found')
    else:
        try:
            return _TYPE_MAP[typekey]
        except KeyError:
            raise TypeError('Decoding type unsupported: {}'.format(typekey))


def _decode_authority(obj, instance):
    """Decodes a termset from a dictionary.

    """
    instance.scopes = [decode(i) for i in obj['scopes']]
    for scope in instance:
        if isinstance(scope, Scope):
            scope.authority = instance


def _decode_scope(obj, instance):
    """Decodes a termset from a dictionary.

    """
    instance.collections = [decode(i) for i in obj['collections']]
    for collection in instance:
        if isinstance(collection, Collection):
            collection.scope = instance


def _decode_collection(obj, instance):
    """Decodes a termset from a dictionary.

    """
    instance.terms = [decode(i) if isinstance(i, dict) else i for i in obj['terms']]
    instance.term_name_regex = obj.get('term_name_regex')
    for term in [instance]:
        if isinstance(term, Term):
            term.collection = instance


def _decode_term(obj, instance):
    """Decodes a term from a dictionary.

    """
    instance.alternative_name = obj.get('alternative_name')
    instance.alternative_url = obj.get('alternative_url')
    instance.associations = obj.get('associations', [])
    instance.idx = obj['idx']
    instance.name_raw = obj.get('name_raw')
    instance.status = obj['status']
    instance.synonyms = obj.get('synonyms', [])
    if instance.parent:
        instance.parent = uuid.UUID(str(obj['parent']))


def _decode_node(obj, instance):
    """Decodes a node instance from a dictionary representation.

    """
    instance.create_date = arrow.get(obj['create_date']).datetime
    instance.data = obj.get('data', dict())
    instance.description = obj.get('description')
    instance.label = obj['label']
    instance.name = obj['name']
    instance.uid = uuid.UUID(str(obj['uid']))
    instance.url = obj.get('url')


# Map of supported types to decoding functions.
_DECODERS = {
    Authority: _decode_authority,
    Collection: _decode_collection,
    Scope: _decode_scope,
    Term: _decode_term
}

# Map of supported type keys to types.
_TYPE_MAP = {
    Authority.__module__: Authority,
    Collection.__module__: Collection,
    Scope.__module__: Scope,
    Term.__module__: Term
}



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._parser.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates parsing of names to match vocabulary entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
from pyessv._archive import load
from pyessv._constants import NODE_TYPE_AUTHORITY
from pyessv._constants import NODE_TYPE_SCOPE
from pyessv._constants import NODE_TYPE_COLLECTION
from pyessv._constants import NODE_TYPE_TERM
from pyessv._exceptions import ParsingError
from pyessv._utils.compat import str



def parse_namespace(namespace, strict=True):
    """Parses a namespace within a vocabulary hierachy.

    :param str namespace: Vocabulary namespace, e.g. wcrp.
    :param bool strict: Flag indicating whether to apply strict lookup rules.

    """
    ns = str(namespace).split(':')
    if len(ns) == 0 or len(ns) > 4:
        raise ValueError('Invalid namespace')

    ns = ns + [None for i in range(4 - len(ns))]

    return parse(ns[0], ns[1], ns[2], ns[3], strict)


def parse(authority, scope=None, collection=None, term=None, strict=True):
    """Parses a name within a vocabulary hierachy.

    :param str authority: Vocabulary authority, e.g. wcrp.
    :param str scope: Vocabulary scope, e.g. global.
    :param str collection: Vocabulary collection, e.g. institute-id.
    :param str term: Vocabulary term, e.g. ipsl.
    :param bool strict: Flag indicating whether to apply strict lookup rules.

    """
    targets = [
        _NodeInfo(NODE_TYPE_AUTHORITY, authority, strict),
        _NodeInfo(NODE_TYPE_SCOPE, scope, strict),
        _NodeInfo(NODE_TYPE_COLLECTION, collection, strict),
        _NodeInfo(NODE_TYPE_TERM, term, strict)
    ]

    for target in [i for i in targets if i.name is not None]:
        node = load(
            targets[0].get_name(target),
            targets[1].get_name(target),
            targets[2].get_name(target),
            targets[3].get_name(target)
            )
        target.set_node(node)

    return target.node.name


class _NodeInfo(object):
    """Information about a node whose name is being parsed.

    """
    def __init__(self, typekey, name, strict):
        """Instance constructor.

        """
        self.node = None
        self.name = name
        self.strict = strict
        self.typekey = typekey


    @property
    def formatted_name(self):
        """Gets formatted name used to search archive.

        """
        name = str(self.name)
        if not self.strict:
            name = name.strip().lower()

        return name


    def get_name(self, target):
        """Gets parsing relative name.

        """
        if self.node:
            return self.node.name
        if target == self:
            return self.formatted_name


    def set_node(self, node):
        """Sets node returned from archive search.

        """
        if node is None:
            raise ParsingError(self.typekey, self.name)
        if self.strict:
            if node.name != self.name:
                raise ParsingError(self.typekey, self.name)
        self.node = node



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.authority.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary authority, e.g. WGCM.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import pyessv
from pyessv._model.node import Node



class Authority(Node):
    """An authority assuming responsibity for governance of vocabularies.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Authority, self).__init__(pyessv.NODE_TYPE_AUTHORITY)

        self.scopes = list()


    def __len__(self):
        """Returns number of items in managed collection.

        """
        return len(self.scopes)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self.scopes,
                           key=lambda i: i if isinstance(i, basestring) else i.name))


    def __getitem__(self, key):
        """Returns a child section item.

        """
        # Match against key.
        comparator = Node.get_comparator(key)
        for item in self.scopes:
            if comparator(item) == key:
                return item


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return []



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._validation.py
   :copyright: Copyright "December 01, 2016', IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates domain model class instance validation.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import datetime
import inspect
import re
import uuid

from pyessv._constants import NODE_TYPE_SET
from pyessv._constants import GOVERNANCE_STATUS_SET
from pyessv._constants import REGEX_CANONICAL_NAME
from pyessv._model import NODE_TYPES
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._model import Term
from pyessv._utils.compat import basestring
from pyessv._utils.compat import urlparse



def is_valid(instance):
    """Gets flag indicating validity status of a domain model class.

    :returns: Validity status of a domain model class.
    :rtype: bool

    """
    return len(validate_node(instance)) == 0


def get_errors(instance):
    """Returns sorted list of domain model class instance validation errors.

    :returns: Sorted list of domain model class instance validation errors.
    :rtype: list

    """
    return sorted(list(validate_node(instance)))


def validate_node(instance):
    """Validates an instance of a domain node.

    :param instance: Sub-class of pyessv.Node

    :returns: Set of instance validation errrors.
    :rtype: set

    """
    if not isinstance(instance, NODE_TYPES):
        raise NotImplementedError('Invalid instance: unknown type')

    validators = _validate_core()
    if isinstance(instance, Authority):
        validators += _validate_authority()
    elif isinstance(instance, Scope):
        validators += _validate_scope()
    elif isinstance(instance, Collection):
        validators += _validate_collection()
    elif isinstance(instance, Term):
        validators += _validate_term()

    errs = set()
    for validator in validators:
        field = validator.__name__.split("_validate_")[-1]
        try:
            validator(instance)
        except AssertionError as err:
            errs.add('{}. Invalid {}'.format(instance.__class__.__name__, field))

    return errs


def _validate_core():
    """Returns common instance validators.

    """
    def _validate_create_date(i):
        assert isinstance(i.create_date, datetime.datetime)

    def _validate_data(i):
        if i.data is not None:
            assert isinstance(i.data, dict)

    def _validate_description(i):
        if isinstance(i, (Authority, Scope, Collection)):
            _assert_string(i.description)
        elif i.description is not None:
            _assert_string(i.description)

    def _validate_label(i):
        _assert_string(i.label)

    def _validate_typekey(i):
        assert i.typekey in NODE_TYPE_SET

    def _validate_uid(i):
        assert isinstance(i.uid, uuid.UUID)

    def _validate_url(i):
        if i.url is not None:
            _assert_url(i.url)

    return [
        _validate_create_date,
        _validate_data,
        _validate_description,
        _validate_label,
        _validate_typekey,
        _validate_uid,
        _validate_url
    ]


def _validate_authority():
    """Returns Authority instance validators.

    """
    def _validate_scopes(i):
        _assert_iterable(i.scopes, Scope)

    def _validate_name(i):
        _assert_string(i.name, REGEX_CANONICAL_NAME)

    return [
        _validate_scopes,
        _validate_name
        ]


def _validate_scope():
    """Returns Scope instance validators.

    """
    def _validate_authority(i):
        assert isinstance(i.authority, Authority)

    def _validate_collections(i):
        _assert_iterable(i.collections, Collection)

    def _validate_name(i):
        _assert_string(i.name, REGEX_CANONICAL_NAME)

    return [
        _validate_authority,
        _validate_collections,
        _validate_name
        ]


def _validate_collection():
    """Returns Collection instance validators.

    """
    def _validate_scope(i):
        assert isinstance(i.scope, Scope)

    def _validate_terms(i):
        _assert_iterable(i.terms, Term)

    def _validate_name(i):
        _assert_string(i.name, REGEX_CANONICAL_NAME)

    return [
        _validate_scope,
        _validate_terms,
        _validate_name
        ]


def _validate_term():
    """Returns Term instance validators.

    """
    def _validate_alternative_name(i):
        if i.alternative_name is not None:
            _assert_string(i.alternative_name)

    def _validate_alternative_url(i):
        if i.alternative_url is not None:
            _assert_url(i.alternative_url)

    def _validate_collection(i):
        assert isinstance(i.collection, Collection)

    def _validate_idx(i):
        assert isinstance(i.idx, int)

    def _validate_name(i):
        if i.collection.term_name_regex is None:
            reg_ex = REGEX_CANONICAL_NAME
        else:
            reg_ex = i.collection.term_name_regex
        _assert_string(i.name, reg_ex)

    def _validate_parent(i):
        if i.parent is not None:
            assert isinstance(i.parent, Term)

    def _validate_status(i):
        assert i.status in GOVERNANCE_STATUS_SET

    def _validate_synonyms(i):
        _assert_iterable(i.synonyms, _assert_string)

    return [
        _validate_alternative_name,
        _validate_alternative_url,
        _validate_collection,
        _validate_idx,
        _validate_name,
        _validate_parent,
        _validate_status,
        _validate_synonyms
        ]


def _assert_string(val, reg_ex=None):
    """Asserts a string value.

    """
    assert isinstance(val, basestring)
    assert len(val.strip()) > 0
    if reg_ex:
        assert re.compile(reg_ex).match(val) is not None


def _assert_url(val):
    """Asserts a url value.

    """
    _assert_string(val)
    url = urlparse(val)
    assert url.netloc and url.scheme


def _assert_iterable(val, modifier):
    """Asserts an iterable value.

    """
    assert isinstance(val, list)
    for i in val:
        if inspect.isfunction(modifier):
            modifier(i)
        else:
            assert isinstance(i, modifier)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._codecs.__init__.py
   :copyright: Copyright "Sep 4, 2013", Earth System Documentation
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encpasulates transformations of terms from one format to another.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._codecs import dictionary
from pyessv._codecs import json
from pyessv._constants import ENCODING_DICT
from pyessv._constants import ENCODING_JSON
from pyessv._model import NODE_TYPES
from pyessv._utils.compat import basestring



# Codecs mapped by encoding.
_CODECS = {
	ENCODING_DICT: dictionary,
	ENCODING_JSON: json
    }

# Map of encodings to allowed input types when decoding.
_DECODE_TYPE_WHITELIST = {
    ENCODING_DICT : (dict, ),
    ENCODING_JSON : (basestring, )
    }


def decode(representation, encoding=ENCODING_JSON):
    """Returns a decoded domain model class instance.

    :param str|dict representation: A domain model class instance representation.
    :param str encoding: A supported encoding (dict|json).

    :returns: A domain model class instance.
    :rtype: object

    """
    if representation is None:
        raise ValueError('Cannot decode a null pointer.')
    if not encoding in _CODECS:
        raise NotImplementedError('Unsupported term encoding :: {0}.'.format(encoding))
    if not isinstance(representation, _DECODE_TYPE_WHITELIST[encoding]):
        err = 'Representation unsupported: must be one of {}.'
        err = err.format(_DECODE_TYPE_WHITELIST[encoding])
        raise TypeError(err)

    return _CODECS[encoding].decode(representation)


def encode(target, encoding=ENCODING_JSON):
    """Returns an encoded domain model class instance|collection.

    :param object|list target: Domain model class instance|collection.
    :param str encoding: A supported encoding (dict|json).

    :returns: Target encoded accordingly.
    :rtype: str|dict|list

    """
    if target is None:
        raise ValueError('Null encoding error')
    if encoding not in _CODECS:
        raise NotImplementedError('Invalid encoding: {}'.format(encoding))

    if isinstance(target, NODE_TYPES):
        encoded = _CODECS[encoding].encode(target)
        if isinstance(encoded, basestring):
            encoded = encoded.strip()
        return encoded



# -*- coding: utf-8 -*-

"""
.. module:: test_parser_name.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv name parsing tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose

import pyessv as LIB
import tests.utils as tu



@nose.with_setup(tu.setup, tu.teardown)
def test_parse():
    """Test parsing of anmes at various levels.

    """
    def _get_config(typeof, name, synonym=None):
        target = synonym or name
        return [
            (typeof, target, name, False),
            (typeof, target, LIB.ParsingError if synonym else name, True),
            (typeof, target.upper(), name, False),
            (typeof, target.upper(), LIB.ParsingError, True),
            (typeof, target.title(), name, False),
            (typeof, target.title(), LIB.ParsingError, True)
        ]


    def _test_parse_names(cfg):
        """Inner test.

        """
        typeof, name, expected, strict = cfg

        try:
            if typeof == 'authority':
                result = LIB.parse(name, strict=strict)
            elif typeof == 'scope':
                result = LIB.parse(tu.TEST_AUTHORITY_NAME, name, strict=strict)
            elif typeof == 'collection':
                result = LIB.parse(tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME, name, strict=strict)
            elif typeof == 'term':
                result = LIB.parse(tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME, tu.TEST_COLLECTION_NAME, name, strict=strict)
        except LIB.ParsingError:
            result = LIB.ParsingError

        assert result == expected, \
               'Parsing error: typeof={}.  name={}.  actual = {}.  expected {}.'.format(typeof, name, result, expected)


    def _test_parse_namespace(cfg):
        """Inner test.

        """
        typeof, name, expected, strict = cfg

        if typeof == 'authority':
            parts = []
        elif typeof == 'scope':
            parts = [tu.TEST_AUTHORITY_NAME]
        elif typeof == 'collection':
            parts = [tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME]
        elif typeof == 'term':
            parts = [tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME, tu.TEST_COLLECTION_NAME]
        parts.append(name)

        try:
            result = LIB.parse_namespace(':'.join(parts), strict)
        except LIB.ParsingError:
            result = LIB.ParsingError

        assert result == expected, \
               'Parsing error: typeof={}.  name={}.  actual = {}.  expected {}.'.format(typeof, name, result, expected)


    for typeof, name, synonym in [
        (LIB.NODE_TYPE_AUTHORITY, tu.TEST_AUTHORITY_NAME, None),
        (LIB.NODE_TYPE_SCOPE, tu.TEST_SCOPE_NAME, None),
        (LIB.NODE_TYPE_COLLECTION, tu.TEST_COLLECTION_NAME, None),
        (LIB.NODE_TYPE_TERM, tu.TEST_TERM_NAME, tu.TEST_TERM_SYNONYMS[0]),
        ]:
        config = _get_config(typeof, name)
        if synonym:
            config += _get_config(typeof, name, synonym)

        for cfg in config:
            desc = 'parse --> {}: {} [strict={}]'.format(cfg[0], cfg[1], cfg[3])
            tu.init(_test_parse_names, desc)
            yield _test_parse_names, cfg


    for typeof, name, synonym in [
        (LIB.NODE_TYPE_AUTHORITY, tu.TEST_AUTHORITY_NAME, None),
        (LIB.NODE_TYPE_SCOPE, tu.TEST_SCOPE_NAME, None),
        (LIB.NODE_TYPE_COLLECTION, tu.TEST_COLLECTION_NAME, None),
        (LIB.NODE_TYPE_TERM, tu.TEST_TERM_NAME, tu.TEST_TERM_SYNONYMS[0]),
        ]:
        config = _get_config(typeof, name)
        if synonym:
            config += _get_config(typeof, name, synonym)

        for cfg in config:
            desc = 'parse namespace --> {}: {} [strict={}]'.format(cfg[0], cfg[1], cfg[3])
            tu.init(_test_parse_namespace, desc)
            yield _test_parse_namespace, cfg



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._codecs.dict.encoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encodes a term to a python dictionary.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._model import Term
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._utils.compat import str



def encode(instance):
    """Encodes an instance of a domain model class as a dictionary.

    :param pyessv.Node instance: A domain model class instance to be encoded as a dictionary.

    :returns: Instance encoded as a simple dictionary.
    :rtype: dict

    :raises TypeError: If instance is a domain model class instance.

    """
    try:
        encoder = _ENCODERS[type(instance)]
    except KeyError:
        raise TypeError('Type encoding unsupported: {}'.format(type(instance)))

    obj = dict()
    _encode_node(instance, obj)
    encoder(instance, obj)

    return obj


def _encode_authority(instance, obj):
    """Encodes a term authority as a dictionary.

    """
    obj['scopes'] = [encode(i) for i in instance.scopes]


def _encode_scope(instance, obj):
    """Encodes a term scope as a dictionary.

    """
    obj['collections'] = [encode(i) for i in instance.collections]


def _encode_collection(instance, obj):
    """Encodes a term collection as a dictionary.

    """
    obj['terms'] = ['{}:{}'.format(i.name, i.uid) for i in instance.terms]
    obj['term_name_regex'] = instance.term_name_regex


def _encode_term(instance, obj):
    """Encodes a term as a dictionary.

    """
    obj['idx'] = instance.idx
    obj['name_raw'] = instance.name_raw
    obj['status'] = instance.status
    if instance.alternative_name is not None:
        obj['alternative_name'] = instance.alternative_name
    if instance.alternative_url is not None:
        obj['alternative_url'] = instance.alternative_url
    if instance.parent is not None:
        obj['parent'] = instance.parent.uid
    if len(instance.associations) > 0:
        obj['associations'] = [i.uid for i in instance.associations]
    if len(instance.synonyms) > 0:
        obj['synonyms'] = instance.synonyms


def _encode_node(instance, obj):
    """Encodes a node instance to a dictionary representation.

    """
    obj['_type'] = str(instance.__module__)
    obj['create_date'] = instance.create_date
    obj['label'] = instance.label
    obj['name'] = instance.name
    obj['uid'] = instance.uid
    if instance.data is not None:
        obj['data'] = instance.data
    if instance.description is not None:
        obj['description'] = instance.description
    if instance.url is not None:
        obj['url'] = instance.url


# Map of supported types to encoding functions.
_ENCODERS = {
    Authority: _encode_authority,
    Collection: _encode_collection,
    Scope: _encode_scope,
    Term: _encode_term,
}



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._factory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates creation of domain model class instances.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import uuid

import arrow

import pyessv

from pyessv._constants import NODE_TYPE_AUTHORITY
from pyessv._constants import NODE_TYPE_COLLECTION
from pyessv._constants import NODE_TYPE_SCOPE
from pyessv._constants import NODE_TYPE_TERM
from pyessv._exceptions import ValidationError
from pyessv._model import NODE_TYPE_KEY_MAP
from pyessv._model import Authority
from pyessv._model import Node
from pyessv._model import Scope
from pyessv._model import Collection
from pyessv._model import Term
from pyessv._parser_template import TemplateParser
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
from pyessv._utils.formatter import format_canonical_name
from pyessv._utils.formatter import format_string
from pyessv._validation import validate_node



def create_authority(name, description, label=None, url=None, create_date=None, data=None):
    """Instantiates, initialises & returns a term authority.

    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.

    :returns: A vocabulary authority, e.g. wcrp.
    :rtype: pyessv.Authority

    """
    return _create_node(
        Authority,
        name,
        description,
        label,
        url,
        create_date,
        data
        )


def create_scope(
    authority,
    name,
    description,
    label=None,
    url=None,
    create_date=None,
    data=None
    ):
    """Instantiates, initialises & returns a term scope.

    :param pyessv.Authority authority: CV authority to which scope is bound.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.

    :returns: A vocabulary scope, e.g. cmip6.
    :rtype: pyessv.Scope

    """
    def _callback(instance):
        instance.authority = authority
        authority.scopes.append(instance)

    return _create_node(
        Scope,
        name,
        description,
        label,
        url,
        create_date,
        data,
        _callback
        )


def create_collection(
    scope,
    name,
    description,
    label=None,
    url=None,
    create_date=None,
    data=None,
    term_name_regex=None
    ):
    """Instantiates, initialises & returns a term collection.

    :param pyessv.Scope scope: CV scope to which collection is bound.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.

    :returns: A vocabulary collection, e.g. insitution-id.
    :rtype: pyessv.Collection

    """
    def _callback(instance):
        instance.scope = scope
        instance.term_name_regex = term_name_regex
        scope.collections.append(instance)

    return _create_node(
        Collection,
        name,
        description,
        label,
        url,
        create_date,
        data,
        _callback
        )


def create_term(
    collection,
    name,
    description=None,
    label=None,
    url=None,
    create_date=None,
    data=None
    ):
    """Instantiates, initialises & returns a term.

    :param pyessv.Collection collection: The collection to which the term belongs.
    :param str name: Canonical name.
    :param str description: Informative description.
    :param str label: Label for UI purposes.
    :param str url: Further information URL.
    :param datetime create_date: Creation date.
    :param dict data: Arbirtrary data.

    :returns: A vocabulary term, e.g. ipsl.
    :rtype: pyessv.Term

    """
    def _callback(instance):
        instance.collection = collection
        instance.idx = len(collection)
        instance.name_raw = format_string(name)
        collection.terms.append(instance)

    return _create_node(
        Term,
        name,
        description,
        label,
        url,
        create_date,
        data,
        callback=_callback
        )


def create_template_parser(template, collections):
    """Instantiates, initialises & returns a template parser.

    :param str template: An expression template.
    :param tuple collections: Collections that the template maps to.

    :returns: A vocabulary expression parser.
    :rtype: pyessv.TemplateParser

    """
    assert isinstance(template, basestring), 'Invalid template'
    assert isinstance(collections, tuple), 'Invalid collections'
    assert len(template) > 0, 'Invalid template'
    assert len(collections) > 0, 'Invalid collections'
    assert len([i for i in collections if not isinstance(i, Collection)]) == 0, 'Invalid collections'
    assert template.count('{}') == len(collections), 'Invalid template'

    return TemplateParser(template, collections)


def _create_node(
    typeof,
    name,
    description,
    label,
    url,
    create_date,
    data,
    callback = None
    ):
    """Instantiates, initialises & returns a node.

    """
    # Set core attributes.
    instance = typeof()
    instance.label = format_string(name)
    instance.name = format_canonical_name(name)
    instance.create_date = create_date or arrow.utcnow().datetime
    instance.data = data
    instance.uid = uuid.uuid4()

    # Set other attributes.
    if description is not None:
        instance.description = format_string(description)
    if url is not None:
        instance.url = format_string(url)

    # Set node specific attributes.
    if callback is not None:
        callback(instance)

    # Validate.
    errors = validate_node(instance)
    if errors:
        raise ValidationError(errors)

    return instance



# -*- coding: utf-8 -*-

"""
.. module:: pyessv.utils.formatter.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates name formatting operations.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._utils.compat import str



def format_canonical_name(name):
    """Formats a canonical name prior to accessing archive.

    """
    if name is not None:
        return str(name).strip() \
                        .replace('_', '-') \
                        .replace(' ', '-') \
                        .lower()


def format_string(val):
    """Formats a simple string.

    """
    if val is not None:
        return str(val).strip()


def format_io_name(name):
    """Formats a simple string.

    """
    if name is not None:
        return str(name).strip() \
                        .replace('_', '-') \
                        .replace(' ', '-') \
                        .lower()



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._exceptions.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package exceptions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class ParsingError(ValueError):
	"""A parsing error raised by the package parser.

	"""
	def __init__(self, typeof, name):
		"""Object constructor.

		"""
		msg = 'A parsing error has occurred: typeof={}, name={}'.format(typeof, name)
		super(ParsingError, self).__init__(msg)


class TemplateParsingError(ValueError):
	"""A template parsing error raised.

	"""
	def __init__(self, val):
		"""Object constructor.

		"""
		msg = 'A template parsing error has occurred: {}'.format(val)
		super(TemplateParsingError, self).__init__(msg)


class ValidationError(ValueError):
	"""A validation error raised by the package validator.

	"""
	def __init__(self, msg):
		"""Object constructor.

		"""
		super(ValidationError, self).__init__(msg)


class InvalidAssociationError(ValueError):
	"""An error raised an invalid association is declared.

	"""
	def __init__(self, association):
		"""Object constructor.

		"""
		msg = 'Unsupported association: {}'.format(association)
		super(InvalidAssociationError, self).__init__(msg)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._archive.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulates access to archive.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._cache import cache
from pyessv._cache import get_cached
from pyessv._io_manager import write
from pyessv._utils.formatter import format_canonical_name



def load(authority, scope=None, collection=None, term=None):
    """Loads a CV authority from archive.

    :param str authority: Vocabulary authority, e.g. wcrp.
    :param str scope: Vocabulary scope, e.g. global.
    :param str collection: Vocabulary collection, e.g. institute-id.
    :param str term: Vocabulary term, e.g. ipsl.

    """
    # Format names.
    names = [authority, scope, collection, term]
    names = [format_canonical_name(i) for i in names if i is not None]

    # Set authority (JIT loads cache).
    result = get_cached(names[0])
    if result is None:
        return

    # Return last loaded sub-collection.
    try:
        for name in names[1:]:
            result = result[name]
    except KeyError:
        pass
    else:
        return result


def add(authority):
    """Adds an authority to the archive.

    """
    cache(authority)


def save():
    """Persists archive to file system.

    """
    for authority in get_cached():
        write(authority)




# -*- coding: utf-8 -*-

"""
.. module:: pyessv._utils.compay.py
   :copyright: Copyright "Feb 7, 2013", Earth System Documentation
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: python 2 to 3 compatibilty extensions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

import sys

# -------
# Pythons
# -------

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

try:
    import simplejson as json
except (ImportError, SyntaxError):
    # simplejson does not support Python 3.2, it throws a SyntaxError
    # because of u'...' Unicode literals.
    import json

# ---------
# Specifics
# ---------

if is_py2:
    from urllib import quote, unquote, quote_plus, unquote_plus, urlencode, getproxies, proxy_bypass
    from urlparse import urlparse, urlunparse, urljoin, urlsplit, urldefrag
    from urllib2 import parse_http_list
    import cookielib
    from Cookie import Morsel
    from StringIO import StringIO

    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring
    numeric_types = (int, long, float)
    integer_types = (int, long)

elif is_py3:
    from urllib.parse import urlparse, urlunparse, urljoin, urlsplit, urlencode, quote, unquote, quote_plus, unquote_plus, urldefrag
    from urllib.request import parse_http_list, getproxies, proxy_bypass
    from http import cookiejar as cookielib
    from http.cookies import Morsel
    from io import StringIO

    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    numeric_types = (int, float)
    integer_types = (int,)


# -*- coding: utf-8 -*-

"""
.. module:: pyessv._utils.convert.py
   :copyright: Copyright "Feb 7, 2013", Earth System Documentation
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of library conversion functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import datetime
import json
import re
import types
import uuid

import arrow

from pyessv._utils.compat import numeric_types
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str


# Values considered to be abbreviations.
_ABBREVIATIONS = ('id', 'uid', 'uuid')

# Default string encoding.
_UTF8 = 'utf-8'


def str_to_unicode(val):
    """Converts input to a string literal.

    :param object val: value to be converted to a string literal.

    :returns: A string literal.
    :rtype: str

    """
    if val is None:
        return str()
    if isinstance(val, str):
        return val

    val = str(val).decode(_UTF8).strip()
    if not len(val):
        return str()

    return str(val)


def str_to_camel_case(target, separator='_'):
    """Converts passed name to camel case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to camel case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Convert to pascal case.
        s = str_to_pascal_case(target, separator)

        # Preserve initial separator
        if s[0:len(separator)] == separator:
            r += separator
            s = s[len(separator):]

        # Lower case abbreviations.
        if s.lower() in _ABBREVIATIONS:
            r += s.lower()

        # Lower case initial character.
        elif len(s):
            r += s[0].lower()
            r += s[1:]

    return r


def str_to_underscore_case(target):
    """Helper function to convert a from camel case string to an underscore case string.

    :param target: A string for conversion.
    :type target: str

    :returns: A string converted to underscore case, e.g. account_number.
    :rtype: str

    """
    if target is None or not len(target):
        return ''

    r = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', target)
    r = re.sub('([a-z0-9])([A-Z])', r'\1_\2', r)
    r = r.lower()

    return r


def str_to_pascal_case(target, separator='_'):
    """Converts passed name to pascal case.

    :param target: A string to be converted.
    :type target: str

    :param separator: A separator used to split target string into constituent parts.
    :type separator: str

    :returns: The target string converted to pascal case.
    :rtype: str

    """
    r = ''
    if target is not None and len(target):
        # Preserve initial separator
        if target[0:len(separator)] == separator:
            r = separator

        # Iterate string parts.
        s = target.split(separator)
        for s in s:

            # Upper case abbreviations.
            if s.lower() in _ABBREVIATIONS:
                r += s.upper()

            # Upper case initial character.
            elif (len(s) > 0):
                r += s[0].upper()
                if (len(s) > 1):
                    r += s[1:]

    return r


def dict_keys(as_dict, key_formatter=str_to_pascal_case):
    """Converts keys of a dictionary using the passed key formatter.

    :param dict as_dict: A dictionary.
    :param function key_formatter: A dictionary key formatter function pointer.

    :returns: A dictionary with it's keys formatted accordingly.
    :rtype: dict

    """
    if not isinstance(as_dict, dict):
        return as_dict

    result = {}
    for key, val in as_dict.items():
        if isinstance(val, collections.Mapping):
            result[key_formatter(key)] = dict_keys(val, key_formatter)
        elif isinstance(val, list):
            result[key_formatter(key)] = [dict_keys(i, key_formatter) for i in val]
        else:
            result[key_formatter(key)] = val

    return result



# -*- coding: utf-8 -*-

"""
.. module:: test_parser_template.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv template parsing tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose

import pyessv as LIB

import tests.utils as tu


# Templates.
_TEMPLATE_1 = 'ciclad/cmip6/{}/{}/{}/{}/afilename.nc1'
_TEMPLATE_2 = '{}/{}/{}/{}/afilename.nc2'
_TEMPLATE_3 = 'ciclad/cmip6/{}/{}/{}/{}'
_TEMPLATES = (
    _TEMPLATE_1,
    _TEMPLATE_2,
    _TEMPLATE_3
    )

# Map of templates to parsers.
_PARSERS = {
    _TEMPLATE_1: None,
    _TEMPLATE_2: None,
    _TEMPLATE_3: None
}

# Map of expression templates to valid expressions.
_VALID = {
    _TEMPLATE_1: 'ciclad/cmip6/ipsl/dcpp/hadgem3-gc31-ll/dcppc-atl-spg/afilename.nc1',
    _TEMPLATE_2: 'ipsl/dcpp/hadgem3-gc31-ll/dcppc-atl-spg/afilename.nc2',
    _TEMPLATE_3: 'ciclad/cmip6/ipsl/dcpp/hadgem3-gc31-ll/dcppc-atl-spg'
}

# Map of expression templates to invalid expressions.
_INVALID = {
    _TEMPLATE_1: 'ciclad/cmip6/WWW/XXX/YYY/ZZZ/afilename.nc1',
    _TEMPLATE_2: 'WWW/XXX/YYY/ZZZ/afilename.nc2',
    _TEMPLATE_3: 'ciclad/cmip6/WWW/XXX/YYY/ZZZ'
}


def test_expressions():
    """pyessv-tests: expression: valid.

    """
    def _do_positive_test(template, parser):
        parser.parse(_VALID[template])

    @nose.tools.raises(ValueError)
    def _do_negative_test(template, parser):
        parser.parse(_INVALID[template])

    _setup()
    for template in _TEMPLATES:
        tu.init(_do_positive_test, 'parse expression :: {} --> positive'.format(template))
        yield _do_positive_test, template, _PARSERS[template]
        tu.init(_do_negative_test, 'parse expression :: {} --> negative'.format(template))
        yield _do_negative_test, template, _PARSERS[template]


def _setup():
    """Unit test setup.

    """
    # Load vocabs.
    LIB.init()

    # Set collections.
    collections = (
        LIB.load('wcrp', 'cmip6', 'institution-id'),
        LIB.load('wcrp', 'cmip6', 'activity-id'),
        LIB.load('wcrp', 'cmip6', 'source-id'),
        LIB.load('wcrp', 'cmip6', 'experiment-id')
        )

    # Set parsers.
    for template in _TEMPLATES:
        _PARSERS[template] = LIB.create_template_parser(template, collections)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.term.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary term, e.g. IPSL.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv
from pyessv._model.node import Node
from pyessv._utils.compat import str



class Term(Node):
    """A vocabulary term.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Term, self).__init__(pyessv.NODE_TYPE_TERM)

        self.alternative_name = None
        self.alternative_url = None
        self.associations = list()
        self.collection = None
        self.idx = None
        self.name_raw = None
        self.parent = None
        self.status = pyessv.GOVERNANCE_STATUS_PENDING
        self.synonyms = list()


    def __contains__(self, key):
        """Instance membership predicate.

        """
        key = str(key).strip().lower()

        return key in self.all_names


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority, self.scope, self.collection]


    @property
    def authority(self):
        """Gets governing authority.

        """
        return self.scope.authority


    @property
    def all_names(self):
        """Returns all term names.

        """
        result = [self.name, self.name_raw, self.alternative_name] + self.synonyms
        result = [t for t in result if t is not None and len(t) > 0]

        return set(sorted(result))


    @property
    def scope(self):
        """Gets governing scope.

        """
        return self.collection.scope



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._codecs.json.decoder.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a term from a JSON text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import datetime
import uuid

from pyessv._codecs.dictionary import decoder as dict_decoder
from pyessv._utils import convert
from pyessv._utils.compat import json
from pyessv._utils.compat import str



# ISO date formats.
_ISO_DATE_FORMATS = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%dT%H:%M:%S']

# Default string encoding.
_UTF8 = 'utf-8'



def decode(as_json):
    """Decodes a document from a UTF-8 encoded json text blob.

    :param str as_xml: Term JSON representation.

    :returns: A term instance.
    :rtype: pyessv.Term

    """
    # Decode raw text blob.
    as_json = _decode_blob(as_json)

    # Convert to dictionary.
    as_dict = _JSONDecoder(convert.str_to_underscore_case).decode(as_json)

    # Decode from dictionary.
    return dict_decoder.decode(as_dict)


class _JSONDecoder(json.JSONDecoder):
    """Extends json decoder so as to handle extended types.

    """
    def __init__(self, key_formatter, to_namedtuple=False):
        """Instance constructor.

        """
        json.JSONDecoder.__init__(self, object_hook=self.dict_to_object)
        self.key_formatter = key_formatter
        self.to_namedtuple = to_namedtuple
        self.value_parsers = [
            self._to_datetime,
            self._to_uuid
            ]


    def dict_to_object(self, d):
        """Converts a dictionary to an object.

        """
        # Parse values.
        for k, v in d.items():
            for parser in self.value_parsers:
                if parser(d, k, v):
                    break

        # Format keys.
        if self.key_formatter is not None:
            d = convert.dict_keys(d, self.key_formatter)

        # Return dictionary | named tuple.
        return d


    def _to_datetime(self, d, k, v):
        """Converts a value to datetime.

        """
        if isinstance(v, str) and len(v):
            try:
                float(v)
            except ValueError:
                for format in _ISO_DATE_FORMATS:
                    try:
                        v = datetime.datetime.strptime(v, format)
                    except (ValueError, TypeError):
                        pass
                    else:
                        d[k] = v
                        return True

        return False


    def _to_uuid(self, d, k, v):
        """Converts a value to uuid.UUID.

        """
        if isinstance(v, basestring) and len(v):
            try:
                v = uuid.UUID(v)
            except ValueError:
                pass
            else:
                d[k] = v
                return True

        return False


def _decode_blob(val):
    """Converts input to a string literal.

    :param object val: value to be converted to a string literal.

    :returns: A string literal.
    :rtype: str

    """
    if val is None:
        return str()
    if isinstance(val, str):
        return val

    val = str(val).decode(_UTF8).strip()
    if not len(val):
        return str()

    return str(val)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.__init__.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Model of vocabulary entities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import NODE_TYPE_AUTHORITY
from pyessv._constants import NODE_TYPE_COLLECTION
from pyessv._constants import NODE_TYPE_SCOPE
from pyessv._constants import NODE_TYPE_TERM
from pyessv._model.authority import Authority
from pyessv._model.collection import Collection
from pyessv._model.node import Node
from pyessv._model.scope import Scope
from pyessv._model.term import Term



# Set of supported node types.
NODE_TYPES = (
    Authority,
    Collection,
    Scope,
    Term,
)

# Map of type keys to types.
NODE_TYPE_KEY_MAP = {
    Authority: NODE_TYPE_AUTHORITY,
    Collection: NODE_TYPE_COLLECTION,
    Scope: NODE_TYPE_SCOPE,
    Term: NODE_TYPE_TERM
}



# -*- coding: utf-8 -*-

"""
.. module:: test_cache.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv cache tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import nose

import pyessv

from pyessv._cache import cache
from pyessv._cache import uncache
from pyessv._cache import get_cached
import tests.utils as tu


def test_interface():
	"""pyessv-tests: caching: interface.

	"""
	assert inspect.isfunction(cache)
	assert inspect.isfunction(uncache)
	assert inspect.isfunction(get_cached)


def test_get_cached():
	"""pyessv-tests: caching: get_cached

	"""
	assert isinstance(get_cached('wcrp'), pyessv.Authority)


def test_cache():
	"""pyessv-tests: caching: cache

	"""
	cache(tu.create_authority())

	assert isinstance(get_cached(tu.TEST_AUTHORITY_NAME), pyessv.Authority)


def test_uncache():
	"""pyessv-tests: caching: uncache

	"""
	cache(tu.create_authority())
	uncache(tu.TEST_AUTHORITY_NAME)

	assert get_cached(tu.TEST_AUTHORITY_NAME) is None



# -*- coding: utf-8 -*-

"""
.. module:: test_io_manager.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv I/O tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import io
import json
import nose
import os

import pyessv as LIB
from pyessv._io_manager import delete
from pyessv._io_manager import read
from pyessv._io_manager import write
import tests.utils as tu



def test_interface():
    """pyessv-tests: io: interface.

    """
    assert inspect.isfunction(delete)
    assert inspect.isfunction(read)
    assert inspect.isfunction(write)


def test_directory():
    """pyessv-tests: io: directory.

    """
    assert os.path.isdir(LIB.DIR_ARCHIVE)


def test_read():
    """pyessv-tests: io: read.

    """
    authorities = read()
    assert isinstance(authorities, list)
    for authority in authorities:
	    assert isinstance(authority, LIB.Authority)

    dirs = os.listdir(LIB.DIR_ARCHIVE)
    assert len(authorities) == len(dirs)
    assert dirs == [i.name for i in authorities]


@nose.with_setup(tu.setup, tu.teardown)
def test_write():
    """pyessv-tests: io: write.

    """
    authority_dirs = os.listdir(LIB.DIR_ARCHIVE)
    authority_dir = os.path.join(LIB.DIR_ARCHIVE, tu.TEST_AUTHORITY_NAME)
    authority_manifest = os.path.join(authority_dir, 'MANIFEST')
    scope_dir = os.path.join(authority_dir, tu.TEST_SCOPE_NAME)
    collection_dir = os.path.join(scope_dir, tu.TEST_COLLECTION_NAME)
    term_file = os.path.join(collection_dir, tu.TEST_TERM_NAME)

    assert not os.path.isdir(authority_dir)
    assert not os.path.isfile(authority_manifest)
    assert not os.path.isdir(scope_dir)
    assert not os.path.isdir(collection_dir)
    assert not os.path.isfile(term_file)

    write(LIB.load(tu.TEST_AUTHORITY_NAME))
    assert len(os.listdir(LIB.DIR_ARCHIVE)) == len(authority_dirs) + 1

    assert os.path.isdir(authority_dir)
    assert os.path.isfile(authority_manifest)
    assert os.path.isdir(scope_dir)
    assert os.path.isdir(collection_dir)
    assert os.path.isfile(term_file)
    with io.open(authority_manifest, 'r') as fstream:
    	assert isinstance(json.loads(fstream.read()), dict)
    with io.open(term_file, 'r') as fstream:
    	assert isinstance(json.loads(fstream.read()), dict)


@nose.with_setup(tu.setup, tu.teardown)
def test_delete():
    """pyessv-tests: io: delete.

    """
    authority_dir = os.path.join(LIB.DIR_ARCHIVE, tu.TEST_AUTHORITY_NAME)
    scope_dir = os.path.join(authority_dir, tu.TEST_SCOPE_NAME)
    collection_dir = os.path.join(scope_dir, tu.TEST_COLLECTION_NAME)
    term_file = os.path.join(collection_dir, tu.TEST_TERM_NAME)

    write(LIB.load(tu.TEST_AUTHORITY_NAME))

    authority = LIB.load(tu.TEST_AUTHORITY_NAME)
    scope = LIB.load(tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME)
    collection = LIB.load(tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME, tu.TEST_COLLECTION_NAME)
    term = LIB.load(tu.TEST_AUTHORITY_NAME, tu.TEST_SCOPE_NAME, tu.TEST_COLLECTION_NAME, tu.TEST_TERM_NAME)

    delete(term)
    assert not os.path.isfile(term_file)
    delete(collection)
    assert not os.path.isfile(collection_dir)
    delete(scope)
    assert not os.path.isfile(scope_dir)
    delete(authority)
    assert not os.path.isdir(collection_dir)
    assert not os.path.isdir(scope_dir)
    assert not os.path.isdir(authority_dir)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._initializer.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Initializes library.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os

from pyessv._cache import cache
from pyessv._constants import DIR_ARCHIVE
from pyessv._io_manager import read
from pyessv._utils import logger



def init():
	"""Library initializer.

	"""
	# Verify archive folder exists.
	if not os.path.isdir(DIR_ARCHIVE):
		raise EnvironmentError('{} directory does not exists'.format(DIR_ARCHIVE))

	# Read set of authorities from file system & cache.
	logger.log('Loading vocabularies from {}:'.format(DIR_ARCHIVE))
	for authority in read():
		logger.log('... loaded: {}'.format(authority))
		cache(authority)



# -*- coding: utf-8 -*-

"""
.. module:: test_interface.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv interface tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import pyessv as LIB
import tests.utils as tu



# Set of classes exposed by library.
_CLASSES = {
    'Authority',
    'Collection',
    'Scope',
    'Term'
    }

# Set of constants exposed by library.
_CONSTANTS = {
    'DIR_ARCHIVE',
    'ENCODING_DICT',
    'ENCODING_JSON',
    'NODE_TYPE_AUTHORITY',
    'NODE_TYPE_COLLECTION',
    'NODE_TYPE_SCOPE',
    'NODE_TYPE_TERM',
    'GOVERNANCE_STATUS_ACCEPTED',
    'GOVERNANCE_STATUS_DEPRECATED',
    'GOVERNANCE_STATUS_PENDING',
    'GOVERNANCE_STATUS_REJECTED'
    }

# Set of exceptions exposed by library.
_EXCEPTIONS = {
    'InvalidAssociationError',
    'ParsingError',
    'ValidationError'
    }

# Set of functions exposed by library.
_FUNCS = {
    # ... archive
    'add',
    'load',
    'save',
    # ... codecs
    'decode',
    'encode',
    # ... factory
    'create_authority',
    'create_collection',
    'create_scope',
    'create_term',
    # ... governance
    'accept',
    'deprecate',
    'destroy',
    'reject',
    'reset',
    # ... initialisation
    'init',
    # ... parsing
    'create_template_parser',
    'parse',
    'parse_namespace',
    # ... validation
    'get_errors',
    'is_valid',
    'validate'
}


def test_library_exports():
    """Test set of exports exposed by library.

    """
    def _test_member(member, member_type):
        """Test that library exposes the named member.

        """
        assertor = getattr(tu, 'assert_has_{}'.format(member_type))
        assertor(LIB, member)


    for members, member_type, in (
        (_CLASSES, 'class'),
        (_CONSTANTS, 'constant'),
        (_EXCEPTIONS, 'exception'),
        (_FUNCS, 'function'),
        ):
        for member in sorted(members):
            desc = 'library exposes {} --> {}'.format(member_type, member)
            tu.init(_test_member, desc)
            yield _test_member, member, member_type




# -*- coding: utf-8 -*-

"""
.. module:: test_codecs.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv encoding / decoding tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import nose

import pyessv as LIB
from pyessv._codecs import decode
from pyessv._codecs import encode
from pyessv._utils.compat import basestring
from pyessv._utils.compat import str
import tests.utils as tu



# Types of representation by encoding.
_ENCODING_REPRESENTATION_TYPE = {
    LIB.ENCODING_DICT: dict,
    LIB.ENCODING_JSON: basestring,
}


def test_interface():
    """pyessv-tests: codecs: interface.

    """
    assert inspect.isfunction(decode)
    assert inspect.isfunction(encode)


@nose.with_setup(None, tu.teardown)
def test_encode():
    """pyessv-tests: encode.

    """
    def _test(func, typeof, encoding):
        """Inner test.

        """
        instance = func()
        assert isinstance(instance, typeof)

        representation = encode(instance, encoding)
        assert isinstance(representation, _ENCODING_REPRESENTATION_TYPE[encoding])

        if encoding == LIB.ENCODING_DICT:
            for k, v in representation.iteritems():
                if not k.startswith('_'):
                    assert getattr(instance, k) == v, 'encode fail: {} :: {} :: {}'.format(k, getattr(instance, k), v)


    for func, typeof in (
        (tu.create_authority, LIB.Authority),
        (tu.create_scope, LIB.Scope),
        (tu.create_collection, LIB.Collection),
        (tu.create_term, LIB.Term)
        ):
        desc = 'encode --> {} --> dict'.format(str(typeof).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, func, typeof, LIB.ENCODING_DICT

        desc = 'encode --> {} --> json'.format(str(typeof).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, func, typeof, LIB.ENCODING_JSON



@nose.with_setup(None, tu.teardown)
def test_decode():
    """pyessv-tests: decode.

    """
    def _test(func, typeof, encoding):
        """Inner test.

        """
        instance = func()
        representation = encode(instance, encoding)
        decoded = decode(representation, encoding)
        assert isinstance(decoded, typeof)


    for func, typeof in (
        (tu.create_authority, LIB.Authority),
        (tu.create_scope, LIB.Scope),
        (tu.create_collection, LIB.Collection),
        (tu.create_term, LIB.Term)
        ):
        desc = 'decode --> {} --> dict'.format(str(typeof).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, func, typeof, LIB.ENCODING_DICT

        desc = 'decode --> {} --> json'.format(str(typeof).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, func, typeof, LIB.ENCODING_JSON



# -*- coding: utf-8 -*-

"""
.. module:: test_validation.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv validation tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyessv as LIB
import tests.utils as tu



# Node level test information.
_NODE_TEST_INFO = [
    ('create_date', arrow.utcnow().datetime, ('', '  ', 123)),
    ('data', {'a': 1}, ('', '  ', 123)),
    ('description', tu.TEST_AUTHORITY_DESCRIPTION, ('', '  ')),
    ('name', tu.TEST_AUTHORITY_NAME, (None, '', '  ', 'invalid $#$#$ name')),
    ('url', tu.TEST_AUTHORITY_URL, ('', '  ', 'an-invalid-url')),
    ]


# Test information mapped by node type.
_TEST_INFO = {
    LIB.NODE_TYPE_AUTHORITY: _NODE_TEST_INFO + [
        ('scopes', [], [None, '', '  ', [123]])
    ],
    LIB.NODE_TYPE_SCOPE: _NODE_TEST_INFO + [
        ('collections', [], [None, '', '  ', [123]])
    ],
    LIB.NODE_TYPE_COLLECTION: _NODE_TEST_INFO + [
        ('terms', [], [None, '', '  ', [123]])
    ],
    LIB.NODE_TYPE_TERM: _NODE_TEST_INFO + [
        ('alternative_name', tu.TEST_AUTHORITY_ALTERNATIVE_NAME, ('', '  ')),
        ('alternative_url', tu.TEST_AUTHORITY_ALTERNATIVE_URL, ('', '  ')),
        ('idx', 1, ('', '  ', [123])),
        ('status', LIB.GOVERNANCE_STATUS_PENDING, ('', '  ', [123])),
        ('synonyms', tu.TEST_TERM_SYNONYMS, ('', '  ', [123])),
    ],
}


def test_node():
    """Tests node validation.

    """
    for typeof, factory in (
        (LIB.NODE_TYPE_AUTHORITY, tu.create_authority),
        (LIB.NODE_TYPE_SCOPE, tu.create_scope),
        (LIB.NODE_TYPE_COLLECTION, tu.create_collection),
        (LIB.NODE_TYPE_TERM, tu.create_term)
        ):
        for attr, valid, invalid in _TEST_INFO[typeof]:
            tu.init(_test_node_attr, 'validate --> {}: {}'.format(typeof, attr))
            yield _test_node_attr, factory, attr, valid, invalid


def _test_node_attr(factory, attr, valid_value, invalid_values):
    """Tests node attribute validation.

    """
    instance = factory()
    for value in invalid_values:
        setattr(instance, attr, value)
        assert LIB.is_valid(instance) == False
        assert len(LIB.get_errors(instance)) >= 1, (LIB.get_errors(instance), attr, value)
    setattr(instance, attr, valid_value)
    assert LIB.is_valid(instance) == True, (LIB.get_errors(instance), attr, valid_value)
    assert len(LIB.get_errors(instance)) == 0


def test_regex_collection():
    """pyessv-tests: validate --> reg-ex collection

    """
    collection = tu.create_collection()
    collection.term_name_regex = r'^[a-z\-]*$'
    term = tu.create_term(collection=collection)
    term.name = 'abc-def'
    assert LIB.is_valid(term) == True


def test_regex_collection_negative():
    """pyessv-tests: validate --> reg-ex collection --> negative

    """
    collection = tu.create_collection()
    collection.term_name_regex = r'^[a-z\-]*$'
    term = tu.create_term(collection=collection)
    term.name = 'ABC-DEF'
    assert LIB.is_valid(term) == False



# -*- coding: utf-8 -*-

"""
.. module:: test_authoring.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv authoring tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose
import pyessv

from pyessv._utils.compat import str
import tests.utils as tu



@nose.with_setup(None, tu.teardown)
def test_create():
    """Test creating an authority.

    """
    def _test(factory, typeof):
        """Inner test.

        """
        tu.assert_object(factory(), typeof)


    for factory, typeof in (
        (tu.create_authority, pyessv.Authority),
        (tu.create_scope, pyessv.Scope),
        (tu.create_collection, pyessv.Collection),
        (tu.create_term, pyessv.Term)
        ):
        desc = 'create --> {}'.format(str(typeof).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, factory, typeof



# @nose.with_setup(_setup, tu.teardown)
# def test_delete():
#     """pyessv-tests: authoring: deleting a term.

#     """
#     count = pyessv.get_count()
#     term = tu.get_term()
#     pyessv.delete(term)
#     tu.assert_int(pyessv.get_count(), count - 1)
#     term = tu.get_term()
#     tu.assert_none(term)


# def test_update():
#     """Test updating a term.

#     """
#     def _update_create_date(term):
#         """Update term create date."""
#         term.create_date = tu.get_date()

#     def _update_description(term):
#         """Update term description."""
#         term.description = tu.get_uuid()

#     def _update_domain(term):
#         """Update term domain."""
#         term.domain = tu.get_string(existing=term.domain)

#     def _update_id(term):
#         """Update term id."""
#         term.idx = tu.get_int(existing=term.idx)

#     def _update_kind(term):
#         """Update term kind."""
#         term.kind = tu.get_string(existing=term.kind)

#     def _update_name(term):
#         """Update term name."""
#         term.name = tu.get_string(existing=term.name)

#     def _update_status(term):
#         """Update term status."""
#         term.status = tu.get_string(existing=term.status)

#     def _update_subdomain(term):
#         """Update term subdomain."""
#         term.subdomain = tu.get_string(existing=term.subdomain)

#     def _update_uid(term):
#         """Update term uid."""
#         term.uid = tu.get_uuid()

#     def _assert(term):
#         """Asserts an update."""
#         count = pyessv.get_count()
#         pyessv.save(term)
#         tu.assert_int(pyessv.get_count(), count)
#         term_ = pyessv.get_term(term.domain, term.subdomain, term.kind, term.name)
#         tu.assert_terms(term, term_)

#     @nose.with_setup(_setup, tu.teardown)
#     def _test(update_callback):
#         """Performs update test."""
#         term = tu.get_term()
#         update_callback(term)
#         _assert(term)

#     for func  in (
#         _update_create_date,
#         _update_description,
#         _update_domain,
#         _update_id,
#         _update_kind,
#         _update_name,
#         _update_status,
#         _update_subdomain,
#         _update_uid
#         ):
#         tu.init(_test, 'authoring', inspect.getdoc(func))
#         yield _test, func



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._cache.store.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Manages cache stores.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._cache import store_memory as _memory_store


# Collection of cache stores.
_STORES = (
	_memory_store,
	)


def cache(authority):
    """Caches authority vocabularies.

    """
    for store in _STORES:
    	store.cache(authority)


def get_cached(authority_name=None):
    """Caches authority vocabularies.

    :param str authority_name: Authority name.

    :returns: A pointer to a cached authority.
    :rtype: pyessv.Authority

    """
    return _memory_store.get_cached(authority_name)


def uncache(authority_name):
    """Uncaches authority vocabularies.

    :param str authority_name: Authority name.

    """
    for store in _STORES:
    	store.uncache(authority_name)



# -*- coding: utf-8 -*-

"""
.. module:: json.encoder.py
   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encodes a term to JSON.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import collections
import datetime
import uuid

import arrow

from pyessv._codecs.dictionary import encoder as dict_encoder
from pyessv._utils import convert
from pyessv._utils.compat import json
from pyessv._utils.compat import numeric_types
from pyessv._utils.compat import str



# Set of data types to be ignored when encoding.
_ENCODE_IGNOREABLE = tuple(list(numeric_types) + [type(None), str])

# Set of data types to be converted to string when encoding.
_ENCODE_STRING = (
    basestring,
    arrow.Arrow,
    datetime.datetime,
    uuid.UUID
    )


def encode(instance):
    """Encodes an instance of a domain model class as a JSON text blob.

    :param pyessv.Node instance: A domain model class instance to be encoded as a JSON text blob.

    :returns: Instance encoded as a JSON text blob.
    :rtype: str

    """
    # Convert to dictionary.
    as_dict = dict_encoder.encode(instance)

    # Return JSON.
    return str(dict_to_json(as_dict))


def dict_to_json(as_dict):
    """Converts a dictionary to json.

    :param dict as_dict: A dictionary.

    :returns: A json encoded text blob.
    :rtype: str

    """
    return json.dumps(_to_encodable(as_dict), indent=4, sort_keys=True)


def _to_encodable(obj, key_formatter=convert.str_to_camel_case):
    """Converts data to encodeable representation.

    """
    if isinstance(obj, _ENCODE_IGNOREABLE):
        return obj

    elif isinstance(obj, _ENCODE_STRING):
        return str(obj)

    elif isinstance(obj, collections.Mapping):
        return {str(key_formatter(k)): _to_encodable(v) for k, v in iter(obj.items())}

    elif isinstance(obj, collections.Iterable):
        return [_to_encodable(i) for i in obj]



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._constants.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package constants.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os
import re



# Directory containing vocabulary archive.
DIR_ARCHIVE = os.path.expanduser('~/.esdoc/pyessv-archive')

# Dictionary encoding.
ENCODING_DICT = 'dict'

# JSON encoding.
ENCODING_JSON = 'json'

# Set of supported encodings.
ENCODING_SET = {
    ENCODING_DICT,
    ENCODING_JSON
}

# Node type - an authority governing vocabularies.
NODE_TYPE_AUTHORITY = 'authority'

# Node type - a scope constraining collection of vocabularies.
NODE_TYPE_SCOPE = 'scope'

# Node type - a collection constraining collection of term.
NODE_TYPE_COLLECTION = 'collection'

# Node type - a term.
NODE_TYPE_TERM = 'term'

# Set of allowed name types.
NODE_TYPE_SET = (
  NODE_TYPE_AUTHORITY,
  NODE_TYPE_COLLECTION,
  NODE_TYPE_SCOPE,
  NODE_TYPE_TERM
  )

# Governance state - the term is accepted.
GOVERNANCE_STATUS_ACCEPTED = 'accepted'

# Governance state - the term is obsolete.
GOVERNANCE_STATUS_DEPRECATED = 'obsolete'

# Governance state - the term is pending review.
GOVERNANCE_STATUS_PENDING = 'pending'

# Governance state - the term is rejected.
GOVERNANCE_STATUS_REJECTED = 'rejected'

# Set of allowed governance states.
GOVERNANCE_STATUS_SET = (
    GOVERNANCE_STATUS_PENDING,
    GOVERNANCE_STATUS_ACCEPTED,
    GOVERNANCE_STATUS_REJECTED,
    GOVERNANCE_STATUS_DEPRECATED
)

# Regular expression for validating a canonical name.
REGEX_CANONICAL_NAME = r'^[A-z0-9\_\-\ \.]*$'



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.collection.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary collection, e.g. institute-id.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import pyessv
from pyessv._model.node import Node



class Collection(Node):
    """A vocabulary term collection.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Collection, self).__init__(pyessv.NODE_TYPE_COLLECTION)

        self.scope = None
        self.terms = list()
        self.term_name_regex = None


    def __len__(self):
        """Returns number of terms in managed collection.

        """
        return len(self.terms)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(sorted(self.terms,
                           key=lambda i: i if isinstance(i, basestring) else i.name))


    def __getitem__(self, key):
        """Returns a child section item.

        """
        # Match against key.
        comparator = Node.get_comparator(key)
        for term in self.terms:
            if comparator(term) == key:
                return term

        # Match against a synonym.
        for term in self.terms:
            if key in term.synonyms:
                return term


    def __contains__(self, key):
        """Instance membership predicate.

        """
        return self[key] is not None


    @property
    def ancestors(self):
        """Gets ancestors within archive hierarchy.

        """
        return [self.authority, self.scope]


    @property
    def authority(self):
        """Gets governing authority.

        """
        return self.scope.authority



# -*- coding: utf-8 -*-

"""
.. module:: test_model.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv model tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import nose

import pyessv as LIB

from pyessv._utils.compat import str
import tests.utils as tu



@nose.with_setup(None, tu.teardown)
def test_iterability():
    """Test iterability of domain model.

    """
    def _test(node, count, key):
        """Inner test.

        """
        assert iter(node)
        assert len(node) == count
        assert key in node
        assert node[key] is not None


    _, collection, scope, authority = tu.create_test_entities()
    for node, count, key in (
        (collection, 1, tu.TEST_TERM_NAME),
        (scope, 1, tu.TEST_COLLECTION_NAME),
        (authority, 1, tu.TEST_SCOPE_NAME)
        ):
        desc = 'iterate --> {}'.format(str(type(node)).split('.')[-1][0:-2].lower())
        tu.init(_test, desc)
        yield _test, node, 1, key



# -*- coding: utf-8 -*-

"""
.. module:: test_governance.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyessv governance tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import pyessv

import nose

import tests.utils as tu



@nose.with_setup(tu.setup_and_create_term, tu.teardown)
def _test(governance_action, governance_status):
    """Performs governance tests.

    """
    term = tu.get_term()
    tu.assert_str(term.status, pyessv.GOVERNANCE_STATUS_PENDING)
    governance_action(term)
    tu.assert_str(term.status, governance_status)
    pyessv.save(term)

    tu.setup()
    term = tu.get_term()
    tu.assert_str(term.status, governance_status)


def test():
    """pyessv-tests: governance: accept term

    """


    for func, status, desc in (
        (pyessv.accept, pyessv.GOVERNANCE_STATUS_ACCEPTED, 'accept'),
        # (pyessv.reject, pyessv.GOVERNANCE_STATUS_REJECTED, 'reject'),
        # (pyessv.reset, pyessv.GOVERNANCE_STATUS_PENDING, 'reset'),
        # (pyessv.deprecate, pyessv.GOVERNANCE_STATUS_DEPRECATED, 'deprecate')
        ):
        tu.init(_test, 'governance', '{} term'.format(desc))
        yield _test, func, status



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._cache.memory.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Simple in-memory cache store.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
# Cached loaded vocabulary authorities objects.
_DATA = {}


def cache(authority):
    """Caches authority vocabularies.

    """
    _DATA[authority.name] = authority


def get_cached(authority_name=None):
    """Caches authority vocabularies.

    :param str authority_name: Authority name.

    :returns: A pointer to a cached authority.
    :rtype: pyessv.Authority

    """
    if authority_name is not None:
        try:
            return _DATA[authority_name]
        except KeyError:
            pass
    else:
        return _DATA.values()


def uncache(authority_name):
    """Uncaches authority vocabularies.

    :param str authority_name: Authority name.

    """
    try:
        del _DATA[authority_name]
    except KeyError:
        pass



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.expression_parser.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A vocabulary constrained expression, e.g. a dataset identifier.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import pyessv
from pyessv._model.collection import Collection
from pyessv._exceptions import TemplateParsingError



class TemplateParser(object):
    """A vocabulary template parser.

    """
    def __init__(self, template, collections):
        """Instance constructor.

        """
        if [i for i in collections if not isinstance(i, Collection)]:
            raise ValueError('Invalid collections')
        if template.count('{}') != len(collections):
            raise ValueError('Invalid expression: template & number of collections are mismatched')

        self.template = template
        self.parts = template.split('{}')
        self.collections = collections


    def parse(self, val):
        """Parses a val against a template.

        """
        parsed = ''
        for idx, collection in enumerate(self.collections):
            parsed += self.parts[idx]
            found = False
            for term in collection:
                found = val.startswith(parsed + term.name)
                if found:
                    break
            if found == False:
                raise TemplateParsingError(val)
            parsed += term.name

        parsed += self.parts[idx + 1]
        if not val == parsed:
            raise TemplateParsingError(val)




# -*- coding: utf-8 -*-

"""
.. module:: pyessv._io_manager.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: I/O manager.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import glob
import io
import os
import shutil

from pyessv._codecs import decode
from pyessv._codecs import encode
from pyessv._constants import DIR_ARCHIVE
from pyessv._constants import ENCODING_JSON
from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._model import Term
from pyessv._validation import is_valid



# Manifest file name.
_MANIFEST = 'MANIFEST'


def delete(target):
    """Deletes vocabulary data from file system.

    """
    if not isinstance(target, (Authority, Scope, Collection, Term)):
        raise TypeError()

    elif isinstance(target, Authority):
        action = shutil.rmtree
        io_path = os.path.join(DIR_ARCHIVE, target.io_name)

    elif isinstance(target, Scope):
        action = shutil.rmtree
        io_path = os.path.join(DIR_ARCHIVE, target.authority.io_name)
        io_path = os.path.join(io_path, target.io_name)

    elif isinstance(target, Collection):
        action = shutil.rmtree
        io_path = os.path.join(DIR_ARCHIVE, target.authority.io_name)
        io_path = os.path.join(io_path, target.scope.io_name)
        io_path = os.path.join(io_path, target.io_name)

    elif isinstance(target, Term):
        action = os.remove
        io_path = os.path.join(DIR_ARCHIVE, target.authority.io_name)
        io_path = os.path.join(io_path, target.scope.io_name)
        io_path = os.path.join(io_path, target.collection.io_name)
        io_path = os.path.join(io_path, target.io_name)

    try:
        action(io_path)
    except OSError:
        pass


def read(archive_dir=DIR_ARCHIVE):
    """Reads vocabularies from archive folder (~/.esdoc/pyessv-archive) upon file system.

    :returns: List of vocabulary authorities loaded from archive folder.
    :rtype: list

    """
    return [_read_authority(i) for i in glob.glob('{}/*'.format(archive_dir))]


def _read_authority(dpath):
    """Reads authority CV data from file system.

    :param str dpath: Path to a directory to which an authority's vocabularies have been written.

    :returns: Authority vocabulary data.
    :rtype: pyessv.Authority

    """
    # Directory must exist.
    if not os.path.isdir(dpath):
        raise OSError('Invalid authority directory.')

    # MANIFEST file must exist.
    if not os.path.isfile(os.path.join(dpath, _MANIFEST)):
        raise OSError('Invalid MANIFEST.')

    # Read authority from manifest.
    fpath = os.path.join(dpath, _MANIFEST)
    with open(fpath, 'r') as fstream:
        authority = decode(fstream.read(), ENCODING_JSON)

    # Read terms.
    term_cache = {}
    for scope in authority:
        for collection in scope:
            collection.terms = _read_terms(dpath, scope, collection, term_cache)

    # Set inter-term hierarchies.
    for term in term_cache.values():
        if term.parent in term_cache:
            term.parent = term_cache[term.parent]

    # Set intra-term hierarchies.
    for term in [i for i in term_cache.values() if i.associations]:
        term.associations = [term_cache[i] if i in term_cache else i for i in term.associations]

    return authority


def _read_terms(dpath, scope, collection, term_cache):
    """Reads terms from file system.

    """
    dpath = os.path.join(dpath, scope.io_name)
    dpath = os.path.join(dpath, collection.io_name)
    dpath = os.path.join(dpath, '*')

    return [_read_term(i, collection, term_cache) for i in glob.iglob(dpath)]


def _read_term(fpath, collection, term_cache):
    """Reads terms from file system.

    """
    # Decode term from JSON file.
    with open(fpath, 'r') as fstream:
        term = decode(fstream.read(), ENCODING_JSON)
    term.collection = collection

    term_cache[term.uid] = term

    return term


def write(authority, archive_dir=DIR_ARCHIVE):
    """Writes authority CV data to file system.

    :param pyessv.Authority authority: Authority class instance to be written to file-system.

    """
    # Validate inputs.
    if not isinstance(authority, Authority):
        raise ValueError('Invalid authority: unknown type')
    if not os.path.isdir(archive_dir):
        raise OSError('Invalid authority directory.')
    if not is_valid(authority):
        raise ValueError('Invalid authority: has validation errors')

    # Set directory.
    dpath = os.path.join(archive_dir, authority.io_name)
    try:
        os.makedirs(dpath)
    except OSError:
        pass

    # Write manifest.
    with open(os.path.join(dpath, _MANIFEST), 'w') as fstream:
        fstream.write(encode(authority))

    # Write collections/terms.
    for scope in authority:
        for collection in scope:
            for term in collection:
                _write_term(dpath, term)


def _write_term(dpath, term):
    """Writes a term to the file system.

    """
    # Set directory.
    dpath = os.path.join(dpath, term.scope.io_name)
    dpath = os.path.join(dpath, term.collection.io_name)
    try:
        os.makedirs(dpath)
    except OSError:
        pass

    # Set file path.
    fpath = os.path.join(dpath, term.io_name)

    # Write term JSON file.
    with open(fpath, 'w') as fstream:
        fstream.write(encode(term))



# -*- coding: utf-8 -*-
"""
.. module:: pyessv.__init__.py

   :copyright: @2015 IPSL (http://ipsl.fr)
   :license: GPL / CeCILL
   :platform: Unix
   :synopsis: Python Earth Science Standard Vocabulary library intializer.

.. moduleauthor:: IPSL (ES-DOC) <dev@esdocumentation.org>

"""
__title__ = 'pyessv'
__version__ = '0.3.2.0.0'
__author__ = 'ES-DOC'
__license__ = 'GPL'
__copyright__ = 'Copyright 2017 ES-DOC'



from pyessv._archive import add
from pyessv._archive import load
from pyessv._archive import save

from pyessv._codecs import decode
from pyessv._codecs import encode

from pyessv._constants import DIR_ARCHIVE
from pyessv._constants import ENCODING_DICT
from pyessv._constants import ENCODING_JSON
from pyessv._constants import NODE_TYPE_AUTHORITY
from pyessv._constants import NODE_TYPE_COLLECTION
from pyessv._constants import NODE_TYPE_SCOPE
from pyessv._constants import NODE_TYPE_TERM
from pyessv._constants import GOVERNANCE_STATUS_ACCEPTED
from pyessv._constants import GOVERNANCE_STATUS_DEPRECATED
from pyessv._constants import GOVERNANCE_STATUS_PENDING
from pyessv._constants import GOVERNANCE_STATUS_REJECTED

from pyessv._exceptions import InvalidAssociationError
from pyessv._exceptions import TemplateParsingError
from pyessv._exceptions import ParsingError
from pyessv._exceptions import ValidationError

from pyessv._factory import create_authority
from pyessv._factory import create_collection
from pyessv._factory import create_template_parser
from pyessv._factory import create_scope
from pyessv._factory import create_term

from pyessv._governance import accept
from pyessv._governance import deprecate
from pyessv._governance import destroy
from pyessv._governance import reject
from pyessv._governance import reset

from pyessv._initializer import init

from pyessv._model import Authority
from pyessv._model import Collection
from pyessv._model import Scope
from pyessv._model import Term

from pyessv._parser import parse
from pyessv._parser import parse_namespace

from pyessv._validation import get_errors
from pyessv._validation import is_valid
from pyessv._validation import validate_node as validate


# Auto-initialize.
init()



# -*- coding: utf-8 -*-

"""
.. module:: utils_cv.py

   :copyright: @2013 Earth System Documentation (https://es-doc.org)
   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Exposes other test utility functions.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import inspect
import os
import shutil

import pyessv as LIB
from pyessv._cache import cache
from pyessv._cache import uncache
from tests.utils_assert import assert_objects



# Test authority.
TEST_AUTHORITY = None
TEST_AUTHORITY_NAME = 'test-authority'
TEST_AUTHORITY_DESCRIPTION = 'test-authority-description'
TEST_AUTHORITY_URL = 'https://github.com/ES-DOC/pyesssv-archive/{}'.format(TEST_AUTHORITY_NAME)
TEST_AUTHORITY_ALTERNATIVE_NAME = 'test-authority-alternative-name'
TEST_AUTHORITY_ALTERNATIVE_URL = 'https://github.com/ES-DOC/pyesssv-archive/{}'.format(TEST_AUTHORITY_NAME)

# Test scope.
TEST_SCOPE = None
TEST_SCOPE_NAME = 'test-scope'
TEST_SCOPE_DESCRIPTION = 'test-scope-description'
TEST_SCOPE_URL = '{}/{}'.format(TEST_AUTHORITY_URL, TEST_SCOPE_NAME)

# Test collection.
TEST_COLLECTION = None
TEST_COLLECTION_NAME = 'test-collection'
TEST_COLLECTION_DESCRIPTION = 'test-collection-description'
TEST_COLLECTION_URL = '{}/{}'.format(TEST_SCOPE_URL, TEST_COLLECTION_NAME)

# Test term.
TEST_TERM = None
TEST_TERM_NAME = 'test-term'
TEST_TERM_DESCRIPTION = 'test-term-description'
TEST_TERM_URL = '{}/{}'.format(TEST_COLLECTION_URL, TEST_TERM_NAME)
TEST_TERM_SYNONYMS = ['test-term-synonym-1', 'test-term-synonym-2']


def create_authority():
    """Creates & returns a test authority.

    """
    global TEST_AUTHORITY

    if TEST_AUTHORITY is None:
        TEST_AUTHORITY = LIB.create_authority(
            TEST_AUTHORITY_NAME,
            TEST_AUTHORITY_DESCRIPTION,
            url=TEST_AUTHORITY_URL
            )

    return TEST_AUTHORITY


def create_scope():
    """Creates & returns a test scope.

    """
    global TEST_SCOPE

    if TEST_SCOPE is None:
        TEST_SCOPE = LIB.create_scope(
            TEST_AUTHORITY or create_authority(),
            TEST_SCOPE_NAME,
            TEST_SCOPE_DESCRIPTION,
            url=TEST_SCOPE_URL
            )

    return TEST_SCOPE


def create_collection():
    """Creates & returns a test collection.

    """
    global TEST_COLLECTION

    if TEST_COLLECTION is None:
        TEST_COLLECTION = LIB.create_collection(
            TEST_SCOPE or create_scope(),
            TEST_COLLECTION_NAME,
            TEST_COLLECTION_DESCRIPTION,
            url=TEST_COLLECTION_URL
            )

    return TEST_COLLECTION


def create_term(collection=None):
    """Creates & returns a test term.

    """
    global TEST_TERM

    if TEST_TERM is None:
        TEST_TERM = LIB.create_term(
            collection or TEST_COLLECTION or create_collection(),
            TEST_TERM_NAME,
            description=TEST_TERM_DESCRIPTION,
            url=TEST_TERM_URL
            )
        TEST_TERM.synonyms = TEST_TERM_SYNONYMS

    return TEST_TERM


def create_test_entities():
    """Returns tuple of test entities.

    """
    return \
        create_term(), \
        create_collection(), \
        create_scope(), \
        create_authority()


def init(func, desc=None):
    """Initializes a test function prior to be executed.

    """
    if desc is None:
        desc = inspect.getdoc(func)
    desc = desc.strip()
    if desc[-1] == '.':
        desc = desc[:-1]
    desc = desc[0].lower() + desc[1:]
    func.description = 'pyessv-tests: {}'.format(desc)


def setup():
    """Performs setup functions and then creates a term prior to running a test.

    """
    teardown()
    create_authority()
    create_scope()
    create_collection()
    create_term()
    cache(TEST_AUTHORITY)


def teardown():
    """Performs teardown functions after running a test.

    """
    global TEST_AUTHORITY
    global TEST_SCOPE
    global TEST_COLLECTION
    global TEST_TERM

    uncache(TEST_AUTHORITY)

    TEST_AUTHORITY = None
    TEST_SCOPE = None
    TEST_COLLECTION = None
    TEST_TERM = None

    try:
        shutil.rmtree(os.path.join(LIB.DIR_ARCHIVE, TEST_AUTHORITY_NAME))
    except OSError:
        pass


def assert_terms(term1, term2):
    """Assers equality of 2 terms.

    :param Term term1: A term to be compared against another.
    :param Term term2: A term to be compared against another.

    """
    assert_objects(term1, term2, LIB.Term)
    assert _get_object_attribute_values(term1) == _get_object_attribute_values(term1)


def _get_object_attribute_values(obj):
    """Returns collection of object attribute values.

    """
    return [getattr(obj, k) for k in dir(obj) if k not in dir(obj.__class__)]



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._model.node.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: A node within the pyessv domain model.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import uuid

import arrow

from pyessv._utils.compat import str
from pyessv._utils.formatter import format_io_name



class Node(object):
    """A node within the pyessv domain model.

    """
    def __init__(self, typekey):
        """Instance constructor.

        """
        self.create_date = None
        self.data = None
        self.description = None
        self.label = None
        self.name = None
        self.typekey = typekey
        self.uid = None
        self.url = None


    def __repr__(self):
        """Instance representation.

        """
        return self.namespace


    @property
    def hierarchy(self):
        """Gets hierachy within archive.

        """
        return self.ancestors + [self]


    @property
    def namespace(self):
        """Gets hierachy within archive.

        """
        return ":".join([i.name for i in self.hierarchy])


    @property
    def io_name(self):
        """Returns name formatted for I/O operations.

        """
        return format_io_name(self.name)


    @staticmethod
    def get_comparator(key):
        """Returns an item from managed collection.

        """
        if isinstance(key, int):
            return lambda i: i.idx
        elif isinstance(key, uuid.UUID):
            return lambda i: i.uid
        else:
            key = str(key).strip().lower()
            try:
                uuid.UUID(key)
            except ValueError:
                return lambda i: i.name
            else:
                return lambda i: str(i.uid)



# -*- coding: utf-8 -*-

"""
.. module:: pyessv._governance.py
   :copyright: Copyright "December 01, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Encapsulate govenerance features.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyessv._constants import GOVERNANCE_STATUS_ACCEPTED
from pyessv._constants import GOVERNANCE_STATUS_DEPRECATED
from pyessv._constants import GOVERNANCE_STATUS_PENDING
from pyessv._constants import GOVERNANCE_STATUS_REJECTED
from pyessv._model import Authority
from pyessv._model import Scope
from pyessv._model import Collection
from pyessv._model import Term
from pyessv._model import NODE_TYPES



def accept(target):
    """Marks node as accepted.

    """
    _apply(target, pyessv.GOVERNANCE_STATUS_ACCEPTED)


def deprecate(target):
    """Marks node as deprecated.

    """
    _apply(target, pyessv.GOVERNANCE_STATUS_DEPRECATED)


def destroy(target):
    """Marks node for removal from all persistant state stores.

    """
    _apply(target, pyessv.GOVERNANCE_STATUS_DEPRECATED)


def reject(target):
    """Marks node as rejected.

    """
    _apply(target, pyessv.GOVERNANCE_STATUS_REJECTED)


def reset(target):
    """Resets node status.

    """
    _apply(target, pyessv.GOVERNANCE_STATUS_PENDING)


def _apply(target, status):
	if type(target) not in NODE_TYPES:
		raise TypeError("Cannot apply governance status to a non domain node")

	# Update status.
	target.status = status

	# Cascade.
	children = []
	if isinstance(target, Authority):
		children = target.scopes
	elif isinstance(target, Scope):
		children = target.collections
	elif isinstance(target, Collection):
		children = target.terms
	for child in children:
		_apply(child, status)
